import java.awt.Color;
import java.util.LinkedList;
import java.lang.Math.*;
import javax.swing.*;

/**
 * AiboImage - the image object
 *
 * includes routines for finding blobs given a pixel and color seed
 * these routines are java translations from the dog PAM code plus
 * additional colors, and the only big difference in segmentation 
 * theory is the use of a linked list for the blobs
 *
 */
public class AiboImage {

    // constants, many of these to keep backward compability with dog code
    private final int MAX_BLOBS = 300;
    private final int X_SIZE = 88;
    private final int Y_SIZE = 72;
    private final int NUM_PIX = X_SIZE * Y_SIZE;
    private final int NUM_COLORS = 11;
    private final int NUM_CHANNELS = 8;
    private final int ORANGE = 0;  // These colors are NOT compatible with the canging of color names, values etc
    private final int YELLOW = 1;  // but they are kept in order to keep the java translation of the dog code as
    private final int SBLUE = 2;   // unchanged as possible
    private final int GREEN = 3;
    private final int PINK = 4;
    private final int RED = 5;
    private final int DBLUE = 6;
    private final int CARPET = 7;
    private final int WHITE = 8;
    private final int SILVER = 9;
    private final int BLACK = 10;
    private final int MAX_CONNECTIONS = 500;
    private final int NONE = 255;
    private final int MIN_AREA = 4;

    // variables
    private LinkedList blobs = new LinkedList(); // all the blobs in the image
    private String name; // eg file name
    public int[] yImage, uImage, vImage, threshold;
    private int color;
    private int labelNumber, connectNumber, autoNumber;
    private int[] binImage = new int[NUM_PIX]; // binary image - color information
    private int[] labelledImage = new int[NUM_PIX]; // binary image - labeled image (blobs)
    private int[][] connections = new int[MAX_CONNECTIONS][2]; // connections between blobs

    // native routines from SegmentWrapper.cc
    private native int setImage(int[] yData, int[] uData, int[] vData); // copy the image data to the native segmentation code
    private native int setThreshold(int[] threshold); // set the threshold in the native segmentation code
    private native int[] getSegmentation(int[] chImage); // get native segmentation results

    // Constructor
    public AiboImage(int[] y, int[] u, int[] v, String fileName) {
	this.name = fileName;
	this.yImage = y;
	this.uImage = u;
	this.vImage = v;

	if (CalibrationCore.useNativeSegmentation) {
	    try {

		System.loadLibrary(CalibrationCore.nativeCode); // load c routines
		
		int a = setImage(yImage, uImage, vImage);  // get the image data to the native segmentation object
    
	    }
	    catch (java.lang.UnsatisfiedLinkError e) {
		JOptionPane.showMessageDialog(null, "Error loading native segmentation library\nUsing built in segmentation routines");
		CalibrationCore.useNativeSegmentation = false;
	    }
	}

	clearBinImage(); // clear image color information
    }

    // Updates the blobs for all colors
    public void updateBlobs() {

	int[] chImage = new int[6336];  // only needed for native segmentation

	// throw out the garbage...
	clearBinImage();
	//System.gc();

	for(int i = 0; i<6336; i++) // for all pixels
	    for(int col = 0; col<11; col++) { // for all colors
		if (CalibrationCore.colorTable.isInColorTable(col, yImage[i], uImage[i], vImage[i]) && CalibrationCore.colorsChecked[col]) {
		    int yy = (int)(i / X_SIZE);
		    int xx = (int)(i - (X_SIZE * yy));
		    if (CalibrationCore.useNativeSegmentation) {
			chImage[i] = (col + 1); // need color value to be +1...
			break;
		    }
		    else {
			paintImage(xx, yy, col);
			break;  // this saves the cpu a lot of work....
		    }
		}
	    }

	if (CalibrationCore.useNativeSegmentation) {
	    try {
		setThreshold(CalibrationCore.threshold);
		setImage(yImage, uImage, vImage);
		int[] labeled = getSegmentation(chImage);
		for(int i = 0; i < 6336; i++) {
		    binImage[i] = labeled[i];
		}
	    }
	    catch (java.lang.UnsatisfiedLinkError e) {
		JOptionPane.showMessageDialog(null, "Error loading native segmentation library\nUsing built in segmentation routines");
		CalibrationCore.useNativeSegmentation = false;	
	    }
	}
    }

    /**
     * returns the image fileName - for future use?
     */
    public String getFileName() {
	return name;
    }

    // for development printouts
    private void debug(String a) {
	System.out.println("Debug: " + a);
    }
    
    /**
     * returns the image in RGB representation in a Color[]
     */
    public Color[] getRGB() {

	// to store the RGB data
	Color[] rgbData = new Color[6336];

	int r, g, b;
	int index = 0; // pixel index
	for (int i = 0; i < 88; i++) // x
	    for (int j = 0; j < 72; j++) {  // y

		// Calculate RGB values from YUV
		r = (int)(yImage[index] + 2.028 * (float)(uImage[index] - 128));
		g = (int)(yImage[index] + 1.140 * (float)(vImage[index] - 128));
		b = (int)(yImage[index] - 0.394 * (float)(uImage[index] - 128) - 0.581 * (float)(vImage[index] - 128));

		// make sure color values are in range 0-255
		if (r<0) 
		    r=0;
		else if (r>255) 
		    r=255;
		
		if (g<0) 
		    g=0;
		else if (g>255) 
		    g=255;
		
		if (b<0) 
		    b=0;
		else if (b>255) 
		    b=255;
		   
		// store the color values in the array...
		rgbData[index] = new Color(r,b,g);

		// ...and let's go to the next pixel
		index++;
	    }
	return rgbData;
    }

    /**
     * Returns the image blob representation based on current colortable
     */
    public Color[] getBlob() {
	
	updateBlobs();
	
	Color blacku = Color.gray; // gray background- if no other color for this pixel
	
	// to store the data
	Color[] blobData = new Color[6336];

	for (int i= 0; i<6336; i++) { // for all pixels
	    blobData[i] = blacku; // first, color pixel background gray

	    if (CalibrationCore.useNativeSegmentation) {
		if ((binImage[i] <= 7) && (binImage[i] >= 0)) // native does not handle carpet and up...
		    blobData[i] = CalibrationCore.colorValue[binImage[i]];
	    }
	    else {
		if (binImage[i] != 255) // 255 = no color for this pixel
		    blobData[i] = CalibrationCore.colorValue[binImage[i]];
	    }
	}

	clearBinImage();
	return blobData;
    }

    /**
     * Returns the image blob representation based on current colortable
     * Only used by automatization
     */
    public int[] getBinBlob() {
	updateBlobs();

	// Copy to another array so binImage can be cleared prior to return statement
	int[] blobData = new int[6336];
	for (int i = 0; i < 6336; i++)
	    blobData[i] = binImage[i];
	clearBinImage();

	return blobData;
    }


    /**
     * Selects a pixel in the image, adds that pixels color to the colortable, and returns all the
     * pixels selected in the image, based on colortable, as a boolean[]
     */
    public boolean[] selectPixel(int x, int y) {

	// to store the data
	boolean[] selectedPixels = new boolean[6336];

	this.color = CalibrationCore.color; // color is the user selected color

	int pos = y * X_SIZE + x;
	CalibrationCore.colorTable.addToColorTable(color, yImage[pos], uImage[pos], vImage[pos]);

	for(int i = 0; i<6336; i++) { // for all pixels
	    selectedPixels[i] = false;
	    for(int col = 0; col < 11; col++) // for all colors
		if (CalibrationCore.colorTable.isInColorTable(col, yImage[i], uImage[i], vImage[i]) && CalibrationCore.colorsChecked[col]) {
		    int yy = (int)(i / X_SIZE);
		    int xx = (int)(i - (X_SIZE * yy));
		    paintImage(xx, yy, color); // color the pixel
		    selectedPixels[i] = true;
		}
	}

	clearBinImage();

	return selectedPixels;
    }

    /**
     * Returns the pixels selected in the image, based on current colortable, as a boolean[]
     */
    public boolean[] checkPixelsRGB() {

	// to store the data
	boolean[] selectedPixels = new boolean[6336];

	for(int i = 0; i<6336; i++) { // for all pixels
	    selectedPixels[i] = false;
	    for(int col = 0; col < 11; col++) // for all colors
		if (CalibrationCore.colorTable.isInColorTable(col, yImage[i], uImage[i], vImage[i]) && CalibrationCore.colorsChecked[col]) {
		    int yy = (int)(i / X_SIZE);
		    int xx = (int)(i - (X_SIZE * yy));
		    selectedPixels[i] = true;
		}
	}
		    	
	return selectedPixels;
    }

    // check if blob edge is reached
    private boolean checkEdge(int pos, int newPos, int color) {

	//check Y image
	if (Math.abs(yImage[pos]-yImage[newPos]) > CalibrationCore.threshold[color]) {
	    return false;
	}

	//check U image
	if (Math.abs(uImage[pos]-uImage[newPos]) > CalibrationCore.threshold[color]) {
	    return false;
	}

	//check V image
	if (Math.abs(vImage[pos]-vImage[newPos]) > CalibrationCore.threshold[color]) {
	    return false;
	}

	return true;
    }

    // find blob for given pixel
    private void paintImage(int x, int y, int color) {

	int pos = y * X_SIZE + x;
	int endTable, startTable;
	int newpos;
	int[] xTable = new int[NUM_PIX];
	int[] yTable = new int[NUM_PIX]; // queues needed for labeling

	endTable = 0;
	startTable = -1;

	// check if this pixel is already labeled
	if(binImage[pos] < NUM_COLORS)
	    return;

	// color the pixel
	binImage[pos] = color;

	//take next label value
	labelNumber++ ;

	if (MAX_BLOBS == blobs.size())   { // maximum number of blobs reached
	    labelNumber++ ;
	    return ; // no more blobs
	}

	// create new blob
	Blob blob = new Blob(color, labelNumber);
	blobs.addLast(blob);

	// look for neighbours, until the queue is full
	while (startTable != endTable) {

	    // new position for pixel check
	    pos = y * X_SIZE + x;

	    // label this pixel
	    labelledImage[pos] = labelNumber;

	    updateBlob(blobs.size()-1, x, y); // blobs.size()-1 is index of last blob

	    // check left cell
	    if (x > 0) {
		newpos = pos - 1;
		if((newpos >= 0) && (newpos < NUM_PIX))
		    if(binImage[newpos] > NUM_COLORS) {
			if (checkEdge(pos, newpos, color)) {
			    binImage[newpos] = color;
			    xTable[endTable] = x - 1;
			    yTable[endTable] = y;
			    if (++endTable == NUM_PIX)
				endTable = 0;
			}
		    } 
		    else { // there is already a labeled pixel, check which
			if (binImage[pos] == binImage[newpos])
			    if (labelledImage[newpos] < labelledImage[pos])
				makeConnection(labelledImage[newpos], labelledImage[pos]);
		    }
	    }

	    if(x < (X_SIZE - 1)) {
		// check right cell
		newpos = pos + 1;
		if((newpos >= 0) && (newpos < NUM_PIX))
		    if(binImage[newpos] > NUM_COLORS) {
			if (checkEdge(pos, newpos, color)) {
			    binImage[newpos] = color;
			    xTable[endTable] = x + 1;
			    yTable[endTable] = y;
			    if (++endTable == NUM_PIX)
				endTable = 0;
			}
		    } 
		    else { // there is already a labeled pixel, check which
			if (binImage[pos] == binImage[newpos])
			    if (labelledImage[newpos] < labelledImage[pos]) {
				makeConnection(labelledImage[newpos], labelledImage[pos]);
			    }
		    }
	    }

	    // check upper cell
	    newpos = pos - X_SIZE ;
	    if((newpos >= 0) && (newpos < NUM_PIX))
		if(binImage[newpos] > NUM_COLORS) {
		    if (checkEdge(pos, newpos, color)) {
			binImage[newpos] = color;
			xTable[endTable] = x;
			yTable[endTable] = y - 1;
			if (++endTable == NUM_PIX)
			    endTable = 0;
		    }
		} 
		else { // there is already a labeled pixel, check which
		    if (binImage[pos] == binImage[newpos])
			if (labelledImage[newpos] < labelledImage[pos]) {
			    makeConnection(labelledImage[newpos], labelledImage[pos]);
			}
		}

	    // check lower cell
	    newpos = pos + X_SIZE;
	    if((newpos >= 0) && (newpos < NUM_PIX))
		if(binImage[newpos] > NUM_COLORS) {
		    if (checkEdge(pos, newpos, color)) {
			binImage[newpos] = color;
			xTable[endTable] = x;
			yTable[endTable] = y + 1;
			if (++endTable == NUM_PIX)
			    endTable = 0;
		    }
		} 
		else { // there is already a labeled pixel, check which
		    if (binImage[pos] == binImage[newpos])
			if (labelledImage[newpos] < labelledImage[pos]) {
			    makeConnection(labelledImage[newpos], labelledImage[pos]);
			}
		}

	    startTable++;
	    if (startTable == NUM_PIX)
		startTable = 0;

	    // take next position
	    x = xTable[startTable];
	    y = yTable[startTable];

	} // end of while
    }

    // Connect blobs
    private void makeConnection(int newLabel, int oldLabel) {

	int ind, oldInd, temp;

	oldInd = 0;
	ind = 0;

	while(ind < connectNumber) {
	    if (newLabel == oldLabel) // same blob, no need to connect
		return;

	    for (ind=0; ind < connectNumber; ind++) {
		if(oldLabel == connections[ind][1]) {
		    oldLabel = connections[ind][0];
		    oldInd = ind;
		    if (oldLabel < newLabel) {
			temp = newLabel; // shift oldLabel and newLabel
			newLabel = oldLabel;
			oldLabel = temp;
		    }
		    break;
		}
	    }
	}

	if (newLabel == oldLabel)
	    return;

	// no other connections with oldLabel found
	connections[connectNumber][0] = newLabel;
	connections[connectNumber][1] = oldLabel;

	if (connections[connectNumber][0] < connections[oldInd][0]) {
	    temp = connections[connectNumber][0];
	    connections[connectNumber][0] = connections[oldInd][0];
	    connections[oldInd][0] = temp;
	    temp = connections[connectNumber][1];
	    connections[connectNumber][1] = connections[oldInd][1];
	    connections[oldInd][1] = temp;
	}

	if(++connectNumber == MAX_CONNECTIONS) 
	    connectNumber--;
    }

    private void addBlobs(int newLabel, int oldLabel) {
	int indNew, indOld;

	// add the blob with oldLabel to blob with newlabel
	for(indNew = 0; indNew < blobs.size(); indNew++)
	    if (((Blob)blobs.get(indNew)).getLabel() == newLabel)
		break;
	for(indOld = 0; indOld < blobs.size(); indOld++)
	    if (((Blob)blobs.get(indOld)).getLabel() == oldLabel)
		break ;

	if (indOld == blobs.size()) // reached the end, if 0 is first index in Linkedlist
	    return;

	// add old values
	((Blob)blobs.get(indNew)).setArea(((Blob)blobs.get(indNew)).getArea() + ((Blob)blobs.get(indOld)).getArea());
	((Blob)blobs.get(indNew)).setCummulativeCol(((Blob)blobs.get(indNew)).getCummulativeCol() + ((Blob)blobs.get(indOld)).getCummulativeCol());
	((Blob)blobs.get(indNew)).setCummulativeRow(((Blob)blobs.get(indNew)).getCummulativeRow() + ((Blob)blobs.get(indOld)).getCummulativeRow());

	if (((Blob)blobs.get(indNew)).getMinx() > ((Blob)blobs.get(indOld)).getMinx())
	    ((Blob)blobs.get(indNew)).setMinx(((Blob)blobs.get(indOld)).getMinx());

	if (((Blob)blobs.get(indNew)).getMaxx() < ((Blob)blobs.get(indOld)).getMaxx())
	    ((Blob)blobs.get(indNew)).setMaxx(((Blob)blobs.get(indOld)).getMaxx());

	if (((Blob)blobs.get(indNew)).getMiny() > ((Blob)blobs.get(indOld)).getMiny())
	    ((Blob)blobs.get(indNew)).setMiny(((Blob)blobs.get(indOld)).getMiny());

	if (((Blob)blobs.get(indNew)).getMaxy() < ((Blob)blobs.get(indOld)).getMaxy())
	    ((Blob)blobs.get(indNew)).setMaxy(((Blob)blobs.get(indOld)).getMaxy());

	//delete old blob
	blobs.remove(indOld);
    }

    /**
     * Clear color information for image
     */
    public void clearBinImage() {

	int ind;

	for(ind=0; ind < NUM_PIX; ind++) {
	    binImage[ind] = NONE;
	    labelledImage[ind] = 1000000;
	}

	for(ind=0; ind < blobs.size(); ind++)
	    resetBlob(ind);

	for(ind=0; ind < MAX_CONNECTIONS; ind++) {
	    connections[ind][0] = - 1;
	    connections[ind][1] = - 1;
	}

	connectNumber = 0;
	labelNumber = 0;
    }

    // remove a blob
    private void resetBlob(int ind) {
	blobs.remove(ind);
    }

    // compute the shapes of blobs
    private void computeShapes() {

	int temp, ind;

	for (int i = 0; i < connectNumber; i++)
	    for (int j = i + 1; j < connectNumber; j++)
		if (connections[j][0] < connections[i][0]) {
		    temp = connections[i][0];
		    connections[i][0] = connections[j][0];
		    connections[j][0] = temp;

		    temp = connections[i][1];
		    connections[i][1] = connections[j][1];
		    connections[j][1] = temp;
		}

	// Add Blobs which are connected
	for(ind = (connectNumber-1); ind >= 0; ind--)
	    addBlobs(connections[ind][0],connections[ind][1]);

	int area;

	// compute parameters of blobs
	for(ind = 0; ind < blobs.size(); ind++) {
	    area = ((Blob)(blobs.get(ind))).getArea();
	    if (area < MIN_AREA) // sort out small blobs
		continue;
	}
    }

    // update the blob info
    private void updateBlob(int ind, int row, int col) {

	Blob blob = ((Blob)blobs.get(ind)); // fetch the right blob

	blob.setArea(blob.getArea() + 1);
	blob.setCummulativeCol(blob.getCummulativeCol() + 1);
	blob.setCummulativeRow(blob.getCummulativeRow() + 1);

	if (blob.getMiny() > col)
	    blob.setMiny(col);

	if (blob.getMaxy() < col)
	    blob.setMaxy(col);

	if (blob.getMinx() > row)
	    blob.setMinx(row);

	if (blob.getMaxx() < row)
	    blob.setMaxx(row);
    }



    // inner class to represent blob objects
    private class Blob {

	private int color, label;
	private int area, cummulativeCol, cummulativeRow;
	private int minx, maxx, miny, maxy;

	// consumes A LOT of memory in worst case (eg many blobs)... fix if possible !!!
	int[] blobPixels = new int[6336];

	Blob(int col, int lab){
	    this.color = col;
	    this.label = lab;

	    this.area = 0;
	    this.cummulativeCol = 0;
	    this.cummulativeRow = 0;

	    for (int i = 0; i<(blobPixels.length); i++)
		blobPixels[i] = 0;
	}

	void addPixel(int x, int y) {
	    int index = x + X_SIZE * y;
	    blobPixels[index] = color;
	}

	int[] getPixels() {
	    return blobPixels;
	}

	int getColor() {
	    return color;
	}

	int getLabel() {
	    return label;
	}

	int getArea() {
	    return area;
	}

	int getCummulativeCol() {
	    return cummulativeCol;
	}

	int getCummulativeRow() {
	    return cummulativeRow;
	}

	void setArea(int a) {
	    area = a;
	}

	void setCummulativeCol(int cc) {
	    cummulativeCol = cc;
	}

	void setCummulativeRow(int cr) {
	    cummulativeRow = cr;
	}

	int getMaxx() {
	    return maxx;
	}

	int getMinx() {
	    return minx;
	}

	int getMaxy() {
	    return maxy;
	}

	int getMiny() {
	    return miny;
	}

	void setMaxx(int max) {
	    maxx = max;
	}

	void setMinx(int mix) {
	    minx = mix;
	}

	void setMaxy(int may) {
	    maxy = may;
	}

	void setMiny(int miy) {
	    miny = miy;
	}
    } // end class Blob
} // end class AiboImage


