//
// Copyright 2001,2002,2003 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <stdio.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/OUnits.h>
#include <OPENR/core_macro.h>
#include <OMWares/OMGsensorData.h>
#include <OMWares/CdtStruct.h>
#include <OMWares/OMoNetMessage.h>
#include <OMWares/MMM_CommandList.h>
#include <OMWares/MoNetMessageMaker.h>
#include <string.h>
#include <time.h>
#include "Joystick.h"


Joystick::Joystick() : 
  jbehavior_(J_BOOTING), jmotion_(M_UNDEFINED), mmm_(), msgQ_(), 
  headtilt_(0), headpan_(0), smTryAgain_(false), moveHeadUp(false),
  moveHeadDown(false), moveHeadLeft(false), moveHeadRight(false)
{
}

OStatus
Joystick::DoInit(const OSystemEvent& event)
{
    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;
    return oSUCCESS;
}

OStatus
Joystick::DoStart(const OSystemEvent& event)
{
    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;
    OPENR::SetMotorPower(opowerON);
    OPENR::EnableJointGain(oprimitiveID_UNDEF); // Is it neccesary?
    
    // Removes pseudo-randomness
    srand(static_cast<unsigned>(clock()));

    return oSUCCESS;
}

OStatus
Joystick::DoStop(const OSystemEvent& event)
{
    OPENR::SetMotorPower(opowerOFF); // Done in PowerMonitor?
    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;
    return oSUCCESS;
}

OStatus
Joystick::DoDestroy(const OSystemEvent& event)
{
    DELETE_ALL_SUBJECT_AND_OBSERVER;
    return oSUCCESS;
}

OStatus
Joystick::AddToMessageQueue()
{
   OStatus result = msgQ_.AddMoNetMessage(mmm_.ProvideMoNetMessage());
   if (result != oSUCCESS) {
      OSYSLOG1((osyslogERROR,"msgQ_.AddMoNetMessage() FAILED %d", result));
	return oFAIL;
   }
   return oSUCCESS;
}
OStatus
Joystick::AddToMessageQueue(OMoNetMessage message)
{
   OStatus result = msgQ_.AddMoNetMessage(&message);
   if (result != oSUCCESS) {
      OSYSLOG1((osyslogERROR,"msgQ_.AddMoNetMessage() FAILED %d", result));
	return oFAIL;
   }
   return oSUCCESS;
}

void
Joystick::ReadyMotion(const OReadyEvent&)
{
    SendMotion();
}

void
Joystick::NotifyMotion(const ONotifyEvent& event)
{
  OMoNetMessage* result = (OMoNetMessage*)event.Data(0);
  //OSYSPRINT(("OMONET RESULT : motion: %s\n,behavior: %d\n",
  //                result->motion_.s_,jbehavior_));    
 
  // If "recovering from fall" has finished
  if (strcmp(result->motion_.s_, "ReactiveGU") == 0) {
    jmotion_ = M_UNDEFINED;
    jbehavior_ = J_ONLINE;
  }
  
  // Preparation has been made for execution of MoNet command
  if ((jbehavior_ == J_PREP_SM)&&(strcmp(result->motion_.s_, "oStanding") == 0))
    ExecSpecialMotion();
  
  observer[event.ObsIndex()]->AssertReady();
}

void
Joystick::ReadySpecialMotion(const OReadyEvent& event)
{
    if (jbehavior_ == J_BOOTING) {
      jmotion_ = SM_STAND;
      ExecSpecialMotion();
    }
}

void
Joystick::NotifySpecialMotion(const ONotifyEvent& event)
{
  
  MoNetResult* result = (MoNetResult*)event.Data(0);
  //OSYSPRINT(("MONET RESULT : commandID %d status %d posture %d\n",
  //               result->commandID, result->status, result->posture));

  // The recover motion can possibly be set during this motion.
  if (jbehavior_ != J_RECOVER) {
    if (result->commandID != 2) {
      // Always return to stand motion after executed MoNet command
      jmotion_ = SM_STAND;
      jbehavior_ = J_PREP_SM;
      ExecSpecialMotion();
    }
    else if (smTryAgain_) {
      smTryAgain_ = false;
      ExecSpecialMotion();
    }
    else {
      // MoNet motions complete, back to default state
      jmotion_ = M_UNDEFINED;
      jbehavior_ = J_ONLINE;
    }
  }

  observer[event.ObsIndex()]->AssertReady();
}

void
Joystick::ResultGSensor(const ONotifyEvent& event)
{
    OMGsensorMessage* result = (OMGsensorMessage*)event.Data(0);
    
    if (result->IsPosture(FALL_DOWN_FRONT) ||
        result->IsPosture(FALL_DOWN_REAR ) ||
        result->IsPosture(FALL_DOWN_RIGHT) ||
        result->IsPosture(FALL_DOWN_LEFT )) {

      // AIBO has lost it's balance
        if (jmotion_ != M_RECOVERFALL) {
            jmotion_ = M_RECOVERFALL;
            jbehavior_ = J_RECOVER;
            GetUp();
        }
    }

    observer[event.ObsIndex()]->AssertReady();
}

void
Joystick::ResultSensor(const ONotifyEvent& event)
{
  OSensorFrameVectorData* sensor = (OSensorFrameVectorData*)event.Data(0);

  // Example of how to retrieve sensor data
  // int x_headtilt_ = (int)degrees(sensor->GetData(0)->frame[0].value / 1000000.0);
  // int y_headpan_ = (int)degrees(sensor->GetData(1)->frame[0].value / 1000000.0);

  if (jbehavior_ == J_ONLINE) {
    if (moveHeadUp)    headpan_  -=  PAN_GAIN;
    if (moveHeadDown)  headpan_  +=  PAN_GAIN;
    if (moveHeadRight) headtilt_ -= TILT_GAIN;
    if (moveHeadLeft)  headtilt_ += TILT_GAIN;
    
    // To avoid specific headpositions
    if (headpan_  >  25) headpan_  =  25;
    if (headpan_  < -90) headpan_  = -90;
    if (headtilt_ >  90) headtilt_ =  90;
    if (headtilt_ < -90) headtilt_ = -90;
  
    //OSYSPRINT(("Headposition -> pan: %d, tilt: %d\n",headpan_,headtilt_));
    MoveHead(headpan_,headtilt_);
  }
  observer[event.ObsIndex()]->AssertReady();
}

void
Joystick::SendMotion()
{
  // OMoNet motions not allowed in these states
  if ((jbehavior_ == J_BOOTING)||(jbehavior_ == J_EXEC_SM))
    return;
  
  // Retrieve motion from the OMoNet message queue and execute it
  if (subject[sbjSetMotion]->IsReady()) {
    OMoNetMessage* msg =  msgQ_.GetMoNetMessage();
    if (msg) {
      subject[sbjSetMotion]->SetData(msg, sizeof(OMoNetMessage));
      subject[sbjSetMotion]->NotifyObservers();
      msgQ_.DeleteFirstOne();
    }
  }
}

void 
Joystick::MakeDecision(const ONotifyEvent& event)
{
  // If busy with something the buttons & axis will not be checked
  if (jbehavior_ != J_ONLINE) {
    observer[event.ObsIndex()]->AssertReady();
    return;
  }
  int temp_jmotion_ = jmotion_;
  const char *data = (const char *)event.Data(0);
  char command[256]; 
  sprintf(command,"%s",data);

  if (!strncmp(command,"K",1)) { // Keyboardcommand
    int c;
    if (sscanf(command,"K%d",&c)==1) {
      switch (c) {
      case 0  : jmotion_ = M_STOP_E;       break;
      case 1  : jmotion_ = M_TURNLEFT;     break;
      case 2  : jmotion_ = M_BACKWARD;     break;
      case 3  : jmotion_ = M_TURNRIGHT;    break;
      case 4  : jmotion_ = M_LEFT;         break;
      case 5  : jmotion_ = M_STAND;        break;
      case 6  : jmotion_ = M_RIGHT;        break;
      case 7  : jmotion_ = M_LEFTFORWARD;  break;
      case 8  : jmotion_ = M_FORWARD;      break;
      case 9  : jmotion_ = M_RIGHTFORWARD; break;
      case 10 : jmotion_ = M_STAND;        break;
      case 11 : jmotion_ = SM_SIT;         break;
      case 12 : jmotion_ = SM_SLEEP;       break;
      case 13 : jmotion_ = SM_STANDBYE;    break;
      case 14 : jmotion_ = SM_WAV_MTN_BYE1; break;
      case 15 : jmotion_ = SM_WAV_MTN_BYE2; break;
      case 16 : jmotion_ = SM_WAV_MTN_SLEEP;break;
      default : jmotion_ = M_STAND;
	OSYSPRINT(("Joystick::MakeDecision - keyboardcommand error"));
      }
      
      if (temp_jmotion_ != jmotion_)
        Move();
    }
  }
  else if (!strncmp(command,"J",1)) { // Joystickcommand
    int axis[6];
    int button[8];
    if (sscanf(command,"J%d#%d#%d#%d#%d#%d#%d#%d#%d#%d#%d#%d#%d#%d",
	       &axis[0],&axis[1],&axis[2],&axis[3],&axis[4],&axis[5],
	       &button[0],&button[1],&button[2],&button[3],&button[4],&button[5],&button[6],&button[7])==14) {
      
      // First check buttons... (buttons is in order of importance)
      bool checkAxis = true;
      if (button[B_QUIT]) {
	jmotion_ = M_RESET;
	checkAxis = false;
      } 
      else if (button[B_DISENGAGE]) {
	checkAxis = false;
      }
      else if (button[B_RESET]) {
	jmotion_ = M_RESET;
	checkAxis = false;
      }
      else if (button[B_LEFTKICK]) {
	jmotion_ = M_LEFTKICK;
	checkAxis = false;
      }
      else if (button[B_RIGHTKICK]) {
	jmotion_ = M_RIGHTKICK;
	checkAxis = false;
      }
      else if (button[B_SPECIALACTION]) {
	int action =  (rand()%(SMNUM_TO-SMNUM_FROM)) + SMNUM_FROM;
	OSYSPRINT(("Execute MoNet command: %d\n",action));
	jmotion_ = (JoystickMotion)action;
	checkAxis = false;
      }
      if (button[B_TAILSWING])
	SwingTail();
      else
	StopTail();

      // ...then Axis
      if (checkAxis) {

	// The axis to move the head
	moveHeadUp = false;
	moveHeadDown = false;
	moveHeadLeft = false;
	moveHeadRight = false;
	if (axis[A_2LEFTRIGHT] < -J_THRESHOLD)
	  moveHeadLeft = true;
	else if (axis[A_2LEFTRIGHT] > J_THRESHOLD)
	  moveHeadRight = true;
	if (axis[A_2UPDOWN] < -J_THRESHOLD)
	  moveHeadUp = true;
	else if (axis[A_2UPDOWN] > J_THRESHOLD)
	  moveHeadDown = true;

	//if (axis[A_TURN] > J_THRESHOLD) {
	//}
	//else if (axis[A_TURN] < -J_THRESHOLD) {
	//}
	if (axis[A_UPDOWN] > J_THRESHOLD) { // Backward
	  if (axis[A_LEFTRIGHT] < -J_THRESHOLD) // Left
	    jmotion_ = M_TURNLEFT;
	  else if (axis[A_LEFTRIGHT] > J_THRESHOLD) // Right
	    jmotion_ = M_TURNRIGHT;
	  else
	    jmotion_ = M_BACKWARD;
	}
	else if (axis[A_UPDOWN] < -J_THRESHOLD) { // Forward
	  if (axis[A_LEFTRIGHT] < -J_THRESHOLD) // Left
	    jmotion_ = M_LEFTFORWARD;
	  else if (axis[A_LEFTRIGHT] > J_THRESHOLD) // Right
	    jmotion_ = M_RIGHTFORWARD;
	  else
	    jmotion_ = M_FORWARD;
	}
	else if (axis[A_LEFTRIGHT] < -J_THRESHOLD) // Left
	  jmotion_ = M_LEFT;       
	else if (axis[A_LEFTRIGHT] > J_THRESHOLD) // Right
	  jmotion_ = M_RIGHT;      
	else // Axis not over threshold
	  jmotion_ = M_STAND;
      }
      Move();
    } 
    else
      OSYSPRINT(("Invalid command (Joystick::MakeDecision)\n"));
  }
  observer[event.ObsIndex()]->AssertReady();
}

void
Joystick::Move()
{
  //OSYSPRINT(("Move command of JoystickMotion: %d\n",jmotion_));
  if (GetCommand() != -1) { // GetCommand returns -1 if jmotion_ isn't a MoNet command
    PrepareSpecialMotion();
    return;
  }
            
  switch (jmotion_) {
  case  M_STAND        : SendStandMotion(); return;
  case  M_RESET        : SendResetMotion(); return;
  case  M_STOP_E       : mmm_.MakeWalkingCommand(mmmmcStopMoveEmergency, mmmscWalkStop); break;
  case  M_STOP_N       : mmm_.MakeStepWalkingCommand(mmmmcMoveNow, mmmscWalkForward, 1); break;
  case  M_FORWARD      : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkForward);        break;
  case  M_RIGHTFORWARD : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkRightForward);   break;
  case  M_LEFTFORWARD  : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkLeftForward);    break;
  case  M_RIGHT        : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkRight);          break;
  case  M_LEFT         : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkLeft);           break;
  case  M_BACKWARD     : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkBackward);       break;
  case  M_TURNRIGHT    : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkTurnRight);      break;
  case  M_TURNLEFT     : mmm_.MakeWalkingCommand(mmmmcMoveNow, mmmscWalkTurnLeft);       break;
  case  M_RIGHTKICK    : mmm_.MakeKickingCommand(mmmmcMoveNow, mmmscRightKick, 
                                                 radians(headtilt_));                    break; 
  case  M_LEFTKICK     : mmm_.MakeKickingCommand(mmmmcMoveNow, mmmscLeftKick,
                                                 radians(headtilt_));                    break;
  default              : mmm_.MakeWalkingCommand(mmmmcStopMoveEmergency, mmmscWalkStop); 
                                                 OSYSPRINT(("Joystick::Move - error\n"));
  }
  if (AddToMessageQueue() == oSUCCESS)
    SendMotion();
}

void
Joystick::MoveHead(int tilt, int pan, int roll)
{
  mmm_.MakeHeadCommand(mmmmcMoveNow, mmmscHeadMove, radians(tilt), radians(pan), radians(roll));
  
  if (AddToMessageQueue() == oSUCCESS)
    SendMotion();
}

void
Joystick::MoveHead(int tilt, int pan)
{
    MoveHead(tilt,pan,0);
}

void
Joystick::SwingTail(int tilt, int pan)
{
    mmm_.MakeTailCommand(mmmmcMoveNow, mmmscTailSwing,
                          radians(tilt), radians(pan),        
                          (unsigned int)600); // cycletime 600 ms  
                          
    if (AddToMessageQueue() == oSUCCESS)
        SendMotion();
}

void
Joystick::SwingTail()
{
  SwingTail(0,45);      
}

void
Joystick::StopTail()
{
    mmm_.MakeTailCommand(mmmmcMoveNow, mmmscTailMove,
                          radians(25.0), radians(0.0),        
                          (unsigned int)1000);
                          
    if (AddToMessageQueue() == oSUCCESS)
        SendMotion();      
}

void
Joystick::GetUp()
{
    OSYSPRINT(("Recovering from fall!\n"));
    OMoNetMessage message;
    message.command_ = oStopMoveEmergency;
    
    if (AddToMessageQueue(message) != oSUCCESS)
      return;
    message.command_ = oMoveNow;
    message.network_ = oAll;
    strcpy(message.motion_.s_, "ReactiveGU");

    if (AddToMessageQueue(message) == oSUCCESS) 
      SendMotion(); 
}

void
Joystick::SendStandMotion()
{
  OMoNetMessage message;
  message.command_ = oMoveNow;
  message.network_ = oLegs; // Only move legs to stand position
  strcpy(message.motion_.s_, "oStanding");
  
  if (AddToMessageQueue(message) == oSUCCESS)
    SendMotion();
}

void
Joystick::SendResetMotion()
{
  headpan_ = 0;
  headtilt_ = 0;
  OMoNetMessage message;
  message.command_ = oMoveNow;
  message.network_ = oAll; // Move legs, head, tail to stand position
  strcpy(message.motion_.s_, "oStanding");
  
  if (AddToMessageQueue(message) == oSUCCESS)
    SendMotion();
}

void
Joystick::PrepareSpecialMotion()
{
  if ((GetCommand() == -1)||(jbehavior_ != J_ONLINE))
    return;
  SendResetMotion();
  jbehavior_ = J_PREP_SM;
}

void
Joystick::ExecSpecialMotion()
{
  int command = GetCommand();
  if ((command == -1) || ((jbehavior_ != J_PREP_SM)&&(jbehavior_ != J_BOOTING)))
    return;

  //OSYSPRINT(("Executing Special motion command: %d\n",command));
  if (subject[sbjCommand]->IsReady()) {
    MoNetCommand cmd(command);
    subject[sbjCommand]->SetData(&cmd, sizeof(cmd));
    subject[sbjCommand]->NotifyObservers();
    jbehavior_ = J_EXEC_SM;
  }
  else {
    OSYSPRINT(("MoNet not ready. Special Motion not sent\n"));
    smTryAgain_ = true;
  }
}

int
Joystick::GetCommand()
{
  switch (jmotion_) {
  case SM_SLEEP         : return 0;
  case SM_SIT           : return 1;
  case SM_STAND         : return 2;
  case SM_CHEER         : return 100;
  case SM_DANCE1        : return 101;
  case SM_DANCE2        : return 102;
  case SM_DANCE3        : return 111;
  case SM_SITBYE        : return 103;
  case SM_SITCOME       : return 104;
  case SM_SITDANCE      : return 105;
  case SM_SITLOOSE      : return 106;
  case SM_SITPURR       : return 107;
  case SM_SITWIN        : return 108;
  case SM_SLEEPDENY     : return 109;
  case SM_STANDBYE      : return 110;
  case SM_WAV_WIN       : return 112;
  case SM_WAV_LOOSE     : return 113;
  case SM_WAV_MTN_BYE1  : return 114;
  case SM_WAV_MTN_BYE2  : return 115;
  case SM_WAV_MTN_SLEEP : return 116;
  }
  return -1;   
}
